#include "config.h"
#include <X11/Xlib.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "defs.h"
#include "font.h"
#include "pixmap.h"
#include "globals.h"

#ifndef abs
#define abs(a) ((a) < 0 ? (0 - (a)) : a)
#endif

static int FONT_AddRef(LPFONT pFont)
{
  return ((++pFont->dwRefCount));
}

static int FONT_Release(LPFONT pFont)
{
  if ((--pFont->dwRefCount) == 0)
  {
    pFont->Destroy(pFont);
    return 0;
  }
  return pFont->dwRefCount;
}

static void XFONT_Destroy(LPXFONT pFont)
{
#ifdef DEBUG
  printf("XFONT_Destroy, killing the font\n");
#endif
  if (pFont->m_pXfs)
  {
    XFreeFont(pFont->m_pDisp,pFont->m_pXfs);
  }
  
  if (pFont->m_gc)
  {
    XFreeGC(pFont->m_pDisp,pFont->m_gc);
  }

  free(pFont);
}

static void XFONT_GetExtents(LPXFONT pFont,const char *pText,int numChars,LPFONTRECT ret)
{
  if (ret)
  {
    int dir,ascent,descent;
    XCharStruct x;
    
    XTextExtents(pFont->m_pXfs,pText,numChars,&dir,&ascent,&descent,&x);
    
    ret->ascent  = x.ascent;
    ret->descent = x.descent;
    ret->width   = x.width;
    ret->height  = x.ascent + x.descent;
  }
}

static void XFONT_GetCenter(LPXFONT pFont,const char *pText,int numChars,int width,int height,int *x,int *y)
{
  if (x)
  {
    *x = 0;
  }
  
  if (y)
  {
    /* FONTRECT fr;
       XFONT_GetExtents(pFont,pText,numChars,&fr); */
    
    *y = pFont->m_pXfs->ascent + ((height - pFont->m_pXfs->ascent) / 2);
    /* *y = (height + fr.height)/2; */
  }
}

static int XFONT_GetID(LPXFONT pFont)
{
  return pFont->m_pXfs->fid;
}

static int XFONT_GetWidth(LPXFONT pFont,const char *pText,int numChars)
{
  return XTextWidth(pFont->m_pXfs,pText,numChars);
}

static void XFONT_SetColor(LPXFONT pFont,unsigned long uColor)
{
  XSetForeground(pFont->m_pDisp,pFont->m_gc,uColor);
}

static void XFONT_Render(LPXFONT pFont,Drawable wnd,
			   int x,int y,const char *pName,int numChars)
{
#ifdef DEBUG
  printf("Rendering %s\n",pName);
#endif
  XDrawString(pFont->m_pDisp,wnd,pFont->m_gc,x,y,pName,numChars);
}

LPFONT CreateXFont(Display *pDisp,int screen,Window mainWindow,const char *pName,int size)
{
  XGCValues gcv;
  GC gc;
  XFontStruct *xfs;
  LPXFONT pFont = (LPXFONT)malloc(sizeof(XFONT));
  
  if (pFont)
  {
    xfs = XLoadQueryFont(pDisp,pName);
    if (!xfs)
    {
      xfs = XLoadQueryFont(pDisp,"fixed");
    }
    
    gcv.font = xfs->fid;
    gc = XCreateGC(pDisp,mainWindow,GCFont,&gcv);
    
    pFont->base.dwRefCount = 1;
    pFont->base.ascent     = xfs->ascent;
    pFont->base.descent    = xfs->descent;
    
    pFont->base.AddRef     = FONT_AddRef;
    pFont->base.Release    = FONT_Release;
    pFont->base.GetWidth   = (FONT_GETWIDTH)XFONT_GetWidth;
    pFont->base.GetCenter  = (FONT_GETCENTER)XFONT_GetCenter;
    pFont->base.GetExtents = (FONT_GETEXTENTS)XFONT_GetExtents;
    pFont->base.Render     = (FONT_RENDER)XFONT_Render;
    pFont->base.GetID      = (FONT_GETID)XFONT_GetID;
    pFont->base.SetColor   = (FONT_SETCOLOR)XFONT_SetColor;
    pFont->base.Destroy    = (FONT_DESTROY)XFONT_Destroy;
    
    pFont->m_pDisp = pDisp;
    pFont->m_iScreen = screen;
    pFont->m_pXfs = xfs;
    pFont->m_gc = gc;
  }
  return (LPFONT)pFont;
}

#ifdef HAVE_XFT

static void XFTFONT_Destroy(LPXFTFONT pFont)
{
#ifdef DEBUG
  printf("XFTFONT_Destroy\n");
#endif
  if (pFont->m_pFont)
  {
    XftFontClose(pFont->m_pDisp,pFont->m_pFont);
  }
  
  if (pFont->m_pDraw)
  {
    XftDrawDestroy(pFont->m_pDraw);
  }

  free(pFont);
}

static void XFTFONT_GetCenter(LPXFTFONT pFont,const char *pText,int numChars,int width,int height,int *x,int *y)
{
  if (x)
  {
    *x = 0;
  }
  
  if (y)
  {
    *y = pFont->base.ascent + ((height - (pFont->base.ascent + pFont->base.descent)) / 2);
  }
}

static void XFTFONT_GetExtents(LPXFTFONT pFont,const char *pText,int numChars,LPFONTRECT ret)
{
  if (ret)
  {
    XGlyphInfo info;
    XftTextExtents8(pFont->m_pDisp,pFont->m_pFont,(XftChar8*)pText,numChars,&info);
    
    ret->ascent  = info.height;
    ret->descent = 0;
    ret->width   = info.width;
    ret->height  = info.height;
  }
}

static int XFTFONT_GetID(LPXFTFONT pFont)
{
  return -1;
}

static int XFTFONT_GetWidth(LPXFTFONT pFont,const char *pText,int numChars)
{
  XGlyphInfo pinfo;
  XftTextExtents8(pFont->m_pDisp,pFont->m_pFont,(XftChar8*)pText,numChars,&pinfo);
  
  return pinfo.xOff;
}

static void XFTFONT_SetColor(LPXFTFONT pFont,unsigned long uColor)
{
  XColor clr;
  clr.pixel = uColor;

  XQueryColor(pFont->m_pDisp,DefaultColormap(pFont->m_pDisp,pFont->m_iScreen),&clr);
  
  pFont->m_color.pixel = uColor;
  pFont->m_color.color.red = clr.red;
  pFont->m_color.color.green = clr.green;
  pFont->m_color.color.blue  = clr.blue;
  pFont->m_color.color.alpha = -1; 
}

static void XFTFONT_Render(LPXFTFONT pFont,Drawable wnd,
		    int x,int y,const char *pName,int numChars)
{
#ifdef DEBUG
  printf("Rendering %s\n",pName);
#endif
  /* change drawable if we have to */
#if HAS_XFTDRAWDRAWABLE
  if (XftDrawDrawable(pFont->m_pDraw) != wnd)  
#endif
  {
    XftDrawChange(pFont->m_pDraw,wnd);
  }
  /* now render */
  XftDrawString8(pFont->m_pDraw,&pFont->m_color,pFont->m_pFont,x,y,(XftChar8*)pName,numChars);
}

LPFONT CreateXFTFont(Display *pDisp,int screen,Window mainWindow,const char *pName,int size)
{
  LPXFTFONT pFont = (LPXFTFONT)malloc(sizeof(XFTFONT));
  if (pFont)
  {
    XftFont *pxftFont = XftFontOpenName(pDisp,gScreen,pName);
    XftDraw *pDraw = XftDrawCreate(pDisp,mainWindow,
				   DefaultVisual(pDisp,gScreen),
				   DefaultColormap(pDisp,gScreen));
    
  
    pFont->base.dwRefCount = 1;
    pFont->base.ascent     = pxftFont->ascent;
    pFont->base.descent    = pxftFont->descent;
    
    pFont->base.AddRef     = FONT_AddRef;
    pFont->base.Release    = FONT_Release;
    pFont->base.GetWidth   = (FONT_GETWIDTH)XFTFONT_GetWidth;
    pFont->base.GetCenter  = (FONT_GETCENTER)XFTFONT_GetCenter;
    pFont->base.GetExtents = (FONT_GETEXTENTS)XFTFONT_GetExtents;
    pFont->base.Render     = (FONT_RENDER)XFTFONT_Render;
    pFont->base.GetID      = (FONT_GETID)XFTFONT_GetID;
    pFont->base.SetColor   = (FONT_SETCOLOR)XFTFONT_SetColor;
    pFont->base.Destroy    = (FONT_DESTROY)XFTFONT_Destroy;
    
    pFont->m_pDisp = pDisp;
    pFont->m_iScreen = screen;
    pFont->m_pFont = pxftFont;
    pFont->m_pDraw = pDraw;
    memset(&pFont->m_color,0,sizeof(XftColor));
    pFont->m_color.color.alpha = -1;
  }
  return (LPFONT)pFont;
}

#endif
